// SPDX-FileCopyrightText: 2012 Icinga GmbH <https://icinga.com>
// SPDX-License-Identifier: GPL-2.0-or-later

#ifndef I2DB_IDO_H
#define I2DB_IDO_H

/**
 * @defgroup db_ido IDO library
 *
 * The Icinga library implements database-agnostic IDO functionality.
 */

#include "base/i2-base.hpp"

#endif /* I2DB_IDO_H */
